@@ -10,9 +10,15 @@ import datetime
 import logging
 import time
 
+from tables import NoSuchNodeError, NodeError
+from mopet.exceptions import (
+    ExplorationNotFoundError,
+    Hdf5FileNotExistsError,
+    ExplorationExistsError,
+)
+
 
 class Exploration:
-
     RUN_PREFIX = "run_"
 
     ##############################################
@@ -30,11 +36,11 @@ class Exploration:
         num_gpus: int = None,
     ):
         """Defines a parameter exploration of a given `function`.
-        
+
         :param function: Function to evaluate at each run
         :type function: function
         :param explore_params: Exploration parameters (individual) for each run
-        :type explore_params: dict        
+        :type explore_params: dict
         :param default_params: Default (shared) parameters to load for each run, optional, defaults to None
         :type default_params: dict
         :param exploration_name: Name of the run, will create a name if left empty, defaults to None
@@ -62,9 +68,7 @@ class Exploration:
         self.explore_params = copy.deepcopy(explore_params)
 
         if exploration_name is None:
-            exploration_name = "exploration" + datetime.datetime.now().strftime(
-                "_%Y_%m_%d_%HH_%MM_%SS"
-            )
+            exploration_name = "exploration" + datetime.datetime.now().strftime("_%Y_%m_%d_%HH_%MM_%SS")
         self.exploration_name = exploration_name
 
         if hdf_filename is None:
@@ -95,6 +99,8 @@ class Exploration:
         """Start parameter exploration.
 
         TODO: Pass kwargs in run() to the exploration function
+
+        :raises ExplorationExistsError: if exploration with same name already exists in HDF5 file.
         """
         # Initialize ray
         self._init_ray(num_cpus=self.num_cpus, num_gpus=self.num_gpus)
@@ -102,6 +108,9 @@ class Exploration:
         # Create a list of all combinations of parameters from explore_params
         self.explore_params_list = self._cartesian_product_dict(self.explore_params)
 
+        # Initialize hdf storage
+        self._pre_storage_routine()
+
         # -----------------------------
         # Set up all simulations
         # -----------------------------
@@ -129,7 +138,7 @@ class Exploration:
 
             # start all ray jobs and remember the ray object
             # pylint: disable=no-member
-            ray_returns[run_id] = self._ray_remote.remote(self, run_params)
+            ray_returns[run_id] = _ray_remote.remote(self.function, run_params)
 
             # store this runs explore parameters
             self.run_params_dict[run_id] = copy.deepcopy(update_params)
@@ -145,9 +154,6 @@ class Exploration:
         # Reduce and store all results
         # -----------------------------
 
-        # initialize hdf storage
-        self._pre_storage_routine()
-
         # remember the time
         start_time = time.time()
 
@@ -169,17 +175,15 @@ class Exploration:
 
         self._shutdown_ray()
 
-    def load_results(
-        self, filename=None, exploration_name=None, aggregate=True, all=False
-    ):
-        """Load results from previous explorations. This function 
-        will open an HDF file and look for an exploration. It will 
+    def load_results(self, filename=None, exploration_name=None, aggregate=True, all=False):
+        """Load results from previous explorations. This function
+        will open an HDF file and look for an exploration. It will
         create a Pandas `Dataframe` object (accessible through the
         attribute `.df`) with a list of all runs and their parameters.
-        
+
         You can load the exploration results using following parameters:
 
-        - If `aggregate==True`, all scalar results (such as `float` 
+        - If `aggregate==True`, all scalar results (such as `float`
         or `int`) from the exploration will be added to the Dataframe.
         - If `all==True`, then all results, including arrays and other
         types, will be saved in the attribute `.results`. This can take
@@ -196,6 +200,7 @@ class Exploration:
         :type aggregate: bool, optional
         :param all: Load all results into a dictionary available as the attribute `.results`. Can use a lot of RAM, defaults to False
         :type all: bool, optional
+        :raises Hdf5FileNotExistsError: if file with `filename` does not exist.
         """
         if exploration_name is None:
             exploration_name = self.exploration_name
@@ -212,7 +217,7 @@ class Exploration:
     def get_run(self, run_id=None, run_name=None, filename=None, exploration_name=None):
         """Get a single result from a previous exploration. This function
         will load a single result from the HDF file. Use this function
-        if you want to avoid loading all results to memory, which you can 
+        if you want to avoid loading all results to memory, which you can
         do using `.load_results(all=True)`.
 
         Note: This function will open the HDF for reading but will not close
@@ -228,14 +233,13 @@ class Exploration:
         :type filename: str, optional
         :param exploration_name: Name of the exploration to load data from. Previously used exploration_name will be used if not given, defaults to None
         :type exploration_name: str, optional
-        
+
         :return: Results of the run
         :rtype: dict
+        :raises: NoSuchExplorationError if hdf5 file does not contain `exploration_name` group.
         """
         # get result by id or if not then by run_name (hdf_run)
-        assert (
-            run_id is not None or run_name is not None
-        ), "Either use `run_id` or `run_name`."
+        assert run_id is not None or run_name is not None, "Either use `run_id` or `run_name`."
 
         if exploration_name:
             self.exploration_name = exploration_name
@@ -245,41 +249,31 @@ class Exploration:
 
         if not self._hdf_open_for_reading:
             self._open_hdf(filename)
-        run_results_group = self.h5file.get_node("/" + self.exploration_name, "runs")[
-            run_name
-        ]
+
+        try:
+            run_results_group = self.h5file.get_node("/" + self.exploration_name, "runs")[run_name]
+        except NoSuchNodeError:
+            raise ExplorationNotFoundError(
+                "Exploration %s could not be found in HDF file %s".format(self.exploration_name, self.hdf_filename)
+            )
 
         result = self._read_group_as_dict(run_results_group)
         return result
 
     def _cartesian_product_dict(self, input_dict):
         """Returns the cartesian product of the exploration parameters.
-        
+
         :param input_dict: Parameter names and their values to explore
         :type input_dict: dict
         :return: List of dictionaries of all possible combinations
         :rtype: list
         """
-        return [
-            dict(zip(input_dict.keys(), values))
-            for values in itertools.product(*input_dict.values())
-        ]
+        return [dict(zip(input_dict.keys(), values)) for values in itertools.product(*input_dict.values())]
 
     ##############################################
     ## MULTIPROCESSING
     ##############################################
 
-    @ray.remote
-    def _ray_remote(self, params):
-        """This is a ray remote function (see ray documentation). It runs the `function` on each ray worker.
-        
-        :param params: Parameters of the run.
-        :type params: dict
-        :return: ray object
-        """
-        r = self.function(params)
-        return r
-
     def _init_ray(self, num_cpus: int = None, num_gpus: int = None):
         """Initialize ray.
 
@@ -295,8 +289,7 @@ class Exploration:
         assert ray.is_initialized() is True, "Could not initialize ray."
 
     def _shutdown_ray(self):
-        """Shutdown ray.
-        """
+        """Shutdown ray."""
         ray.shutdown()
         assert ray.is_initialized() is False, "Could not shutdown ray."
 
@@ -306,7 +299,7 @@ class Exploration:
 
     def _store_dict_to_hdf(self, group, dict_data):
         """Stores a dictionary into a group of the hdf file.
-        
+
         :param group: group in hdf file to store data in
         :type group: [type]
         :param dict_data: dictionary with data to store
@@ -316,20 +309,29 @@ class Exploration:
             try:
                 self.h5file.create_array(group, r_key, obj=r_val)
             except:
-                logging.warning(
-                    f"Could not store dict entry {r_key} (type: {type(r_val)})"
-                )
+                logging.warning(f"Could not store dict entry {r_key} (type: {type(r_val)})")
 
     def _init_hdf(self):
         """Create hdf storage file and all necessary groups.
+
+        :raises Hdf5FileNotExistsError, ExplorationAlreadyExists
         """
-        self.h5file = tables.open_file(self.hdf_filename, mode="a")
-        self.run_group = self.h5file.create_group("/", self.exploration_name)
+        try:
+            self.h5file = tables.open_file(self.hdf_filename, mode="a")
+        except IOError:
+            raise Hdf5FileNotExistsError("Hdf5 file {} does not exist".format(self.hdf_filename))
+
+        try:
+            self.run_group = self.h5file.create_group("/", self.exploration_name)
+        except NodeError as e:
+            raise ExplorationExistsError(
+                "Exploration with name {} already exists in HDF5 file {}".format(
+                    self.exploration_name, self.hdf_filename
+                )
+            )
 
         # create group in which all data from runs will be saved
-        self.runs_group = self.h5file.create_group(
-            self.h5file.root[self.exploration_name], "runs"
-        )
+        self.runs_group = self.h5file.create_group(self.h5file.root[self.exploration_name], "runs")
 
         if self.default_params is not None:
             # create group in which all default parameters will be saved
@@ -340,28 +342,24 @@ class Exploration:
             self._store_dict_to_hdf(self.default_params_group, self.default_params)
 
         # create group in which exploration parameters will be saved
-        self.explore_params_group = self.h5file.create_group(
-            self.h5file.root[self.exploration_name], "explore_params"
-        )
+        self.explore_params_group = self.h5file.create_group(self.h5file.root[self.exploration_name], "explore_params")
         self._store_dict_to_hdf(self.explore_params_group, self.explore_params)
 
         # create group in which information about this run is saved
         # self.info_group = self.h5file.create_group("/", "info")
 
     def _pre_storage_routine(self):
-        """Routines for preparing the hdf storage.
-        """
+        """Routines for preparing the hdf storage."""
         # initialize the hdf file
         self._init_hdf()
 
     def _post_storage_routine(self):
-        """Routines for closing the hdf storage.
-        """
+        """Routines for closing the hdf storage."""
         self.h5file.close()
 
     def _store_result(self, result_id, ray_object, run_params):
         """Resolves results from the ray object and stores the results.
-        
+
         :param result_id: id of the run
         :type result_id: int
         :param ray_object: ray object
@@ -375,15 +373,13 @@ class Exploration:
         # resolve the ray object and get the returned dictionary from the evaluation function
         result_dict = ray.get(ray_object)
 
-        assert isinstance(
-            result_dict, dict
-        ), f"Returned result must be a dictionary, is `{type(result_dict)}`."
+        assert isinstance(result_dict, dict), f"Returned result must be a dictionary, is `{type(result_dict)}`."
 
         self._store_result_in_hdf(run_result_name, result_dict, run_params)
 
     def _store_result_in_hdf(self, run_result_name, result_dict, run_params):
         """Stores the results of a ray object of a single run and the parameters of the run.
-        
+
         :param run_result_name: Name of the result
         :type run_result_name: str
         :param run_params: Explored parameters of the run
@@ -414,9 +410,7 @@ class Exploration:
         """
         logging.info("Creating new results DataFrame")
         self.explore_params = self._read_explore_params()
-        self.dfResults = pd.DataFrame(
-            columns=self.explore_params.keys(), index=self.run_ids, dtype=object
-        )
+        self.dfResults = pd.DataFrame(columns=self.explore_params.keys(), index=self.run_ids, dtype=object)
         for key, value in self.params.items():
             self.dfResults.loc[key] = value
         return self.dfResults
@@ -426,25 +420,27 @@ class Exploration:
 
         :param filename: Filename of HDF file, defaults to None
         :type filename: str, optional
+        :raises Hdf5FileNotExistsError
         """
         if filename is not None:
             self.hdf_filename = filename
-        assert (
-            self.hdf_filename is not None
-        ), "No hdf filename was given or previously set."
+        assert self.hdf_filename is not None, "No hdf filename was given or previously set."
+
+        try:
+            self.h5file = tables.open_file(self.hdf_filename, mode="r+")
+        except OSError:
+            raise Hdf5FileNotExistsError("Hdf5 file %s does not exist".format(self.hdf_filename))
 
-        self.h5file = tables.open_file(self.hdf_filename, mode="r+")
         self._hdf_open_for_reading = True
         logging.info(f"{self.hdf_filename} opened for reading.")
 
     def close_hdf(self):
-        """Close a previously opened HDF file.
-        """
+        """Close a previously opened HDF file."""
         self.h5file.close()
         self._hdf_open_for_reading = False
         logging.info(f"{self.hdf_filename} closed.")
 
-    def _aggregate_results(self, exploration_name=None):
+    def _aggregate_results(self, exploration_name=None, arrays=True):
         """Go through all results saved in `.results` and store all floats in the results table.
 
         TODO: Direct reading from hdf without having to load it to memory, like in neurolib
@@ -455,13 +451,19 @@ class Exploration:
         """
         nan_value = np.nan
         logging.info("Aggregating scalar results ...")
-        for runId, parameters in tqdm.tqdm(
-            self.dfResults.iterrows(), total=len(self.dfResults)
-        ):
+        for runId, parameters in tqdm.tqdm(self.dfResults.iterrows(), total=len(self.dfResults)):
             result = self.get_run(runId)
             for key, value in result.items():
-                if isinstance(value, float):
+                # we check for the type of the value and
+                # save it to the datafram accordingly
+                if isinstance(value, (float, int)):
                     self.dfResults.loc[runId, key] = value
+                elif isinstance(value, np.ndarray) and arrays == True:
+                    # to save a numpy array, convert column to object type
+                    if key not in self.dfResults:
+                        self.dfResults[key] = None
+                    self.dfResults[key] = self.dfResults[key].astype(object)
+                    self.dfResults.at[runId, key] = value
                 else:
                     self.dfResults.loc[runId, key] = nan_value
         # drop nan columns
@@ -474,6 +476,7 @@ class Exploration:
         :type exploration_name: str, optional
         :param all: Whether to load everything into ram or not, defaults to True
         :type all: bool, optional
+        :raises Hdf5FileNotExistsError, ExplorationNotFoundError
         """
         if exploration_name is None:
             exploration_name = self.exploration_name
@@ -517,11 +520,10 @@ class Exploration:
         for array in group:
             if isinstance(array, tables.array.Array):
                 value = array.read()
-                if not isinstance(value, np.ndarray):
-                    value = np.array(value)
-                # convert 0-dim arrays to floats
-                if value.ndim == 0:
-                    value = np.float(value)
+                # unwrap 0-dim arrays
+                if type(value) is np.ndarray and value.ndim == 0:
+                    value = array.dtype.type(value)
+
                 key = array._v_name
                 return_dict[key] = value
         return return_dict
@@ -546,9 +548,7 @@ class Exploration:
         :return: Dictionary with explored parameters
         :rtype: dict
         """
-        explore_params_group = self.h5file.get_node(
-            "/" + self.exploration_name, "explore_params"
-        )
+        explore_params_group = self.h5file.get_node("/" + self.exploration_name, "explore_params")
         self.explore_params = self._read_group_as_dict(explore_params_group)
         return self.explore_params
 
@@ -566,3 +566,17 @@ class Exploration:
                 return self.dfResults
         else:
             return self._create_df()
+
+
+@ray.remote
+def _ray_remote(function, params):
+    """This is a ray remote function (see ray documentation). It runs the `function` on each ray worker.
+
+    :param function: function to be executed remotely.
+    :type function: callable
+    :param params: Parameters of the run.
+    :type params: dict
+    :return: ray object
+    """
+    r = function(params)
+    return r
